@@ -18,17 +18,30 @@
 #ifndef CURSOR_CHANNEL_H_
 # define CURSOR_CHANNEL_H_
 
-#include "spice.h"
-#include "reds.h"
-#include "red-worker.h"
+#include "common-graphics-channel.h"
 #include "red-parse-qxl.h"
 
+G_BEGIN_DECLS
+
 /**
  * This type it's a RedChannel class which implement cursor (mouse)
  * movements.
  * A pointer to CursorChannel can be converted to a RedChannel.
  */
 typedef struct CursorChannel CursorChannel;
+typedef struct CursorChannelClass CursorChannelClass;
+
+#define TYPE_CURSOR_CHANNEL cursor_channel_get_type()
+
+#define CURSOR_CHANNEL(obj) (G_TYPE_CHECK_INSTANCE_CAST((obj), TYPE_CURSOR_CHANNEL, CursorChannel))
+#define CURSOR_CHANNEL_CLASS(klass) \
+    (G_TYPE_CHECK_CLASS_CAST((klass), TYPE_CURSOR_CHANNEL, CursorChannelClass))
+#define IS_CURSOR_CHANNEL(obj) (G_TYPE_CHECK_INSTANCE_TYPE((obj), TYPE_CURSOR_CHANNEL))
+#define IS_CURSOR_CHANNEL_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE((klass), TYPE_CURSOR_CHANNEL))
+#define CURSOR_CHANNEL_GET_CLASS(obj) \
+    (G_TYPE_INSTANCE_GET_CLASS((obj), TYPE_CURSOR_CHANNEL, CursorChannelClass))
+
+GType cursor_channel_get_type(void) G_GNUC_CONST;
 
 /**
  * Create CursorChannel.
@@ -42,14 +55,15 @@ typedef struct CursorChannel CursorChannel;
  * provided as helper functions and should only be called from the
  * CursorChannel thread.
  */
-CursorChannel*       cursor_channel_new         (RedWorker *worker);
+CursorChannel*       cursor_channel_new         (RedsState *server, QXLInstance *qxl,
+                                                 const SpiceCoreInterfaceInternal *core);
 
 /**
  * Cause the channel to disconnect all clients
  */
 void                 cursor_channel_disconnect  (CursorChannel *cursor);
 void                 cursor_channel_reset       (CursorChannel *cursor);
-void                 cursor_channel_init        (CursorChannel *cursor);
+void                 cursor_channel_do_init     (CursorChannel *cursor);
 void                 cursor_channel_process_cmd (CursorChannel *cursor, RedCursorCmd *cursor_cmd);
 void                 cursor_channel_set_mouse_mode(CursorChannel *cursor, uint32_t mode);
 
@@ -58,7 +72,7 @@ void                 cursor_channel_set_mouse_mode(CursorChannel *cursor, uint32
  * This is the equivalent of RedChannel client connect callback.
  * See comment on cursor_channel_new.
  */
-void                 cursor_channel_connect     (CursorChannel *cursor, RedClient *client, 
+void                 cursor_channel_connect     (CursorChannel *cursor, RedClient *client,
                                                  RedsStream *stream,
                                                  int migrate,
                                                  uint32_t *common_caps, int num_common_caps,
@@ -71,4 +85,6 @@ void                 cursor_channel_connect     (CursorChannel *cursor, RedClien
  */
 void                 cursor_channel_client_migrate(RedChannelClient *client);
 
+G_END_DECLS
+
 #endif /* CURSOR_CHANNEL_H_ */
