@@ -39,6 +39,8 @@
 #include "reds.h"
 #include "reds-stream.h"
 #include "red-channel.h"
+#include "red-channel-client.h"
+#include "red-client.h"
 #include "inputs-channel-client.h"
 #include "main-channel-client.h"
 #include "inputs-channel.h"
@@ -56,6 +58,27 @@
 #define RECEIVE_BUF_SIZE \
     (4096 + (REDS_AGENT_WINDOW_SIZE + REDS_NUM_INTERNAL_AGENT_MESSAGES) * SPICE_AGENT_MAX_DATA_SIZE)
 
+struct InputsChannel
+{
+    RedChannel parent;
+
+    uint8_t recv_buf[RECEIVE_BUF_SIZE];
+    VDAgentMouseState mouse_state;
+    int src_during_migrate;
+    SpiceTimer *key_modifiers_timer;
+
+    SpiceKbdInstance *keyboard;
+    SpiceMouseInstance *mouse;
+    SpiceTabletInstance *tablet;
+};
+
+struct InputsChannelClass
+{
+    RedChannelClass parent_class;
+};
+
+G_DEFINE_TYPE(InputsChannel, inputs_channel, RED_TYPE_CHANNEL)
+
 struct SpiceKbdState {
     uint8_t push_ext_type;
 
@@ -104,21 +127,6 @@ RedsState* spice_tablet_state_get_server(SpiceTabletState *st)
     return st->reds;
 }
 
-struct InputsChannel {
-    RedChannel base;
-    uint8_t recv_buf[RECEIVE_BUF_SIZE];
-    VDAgentMouseState mouse_state;
-    int src_during_migrate;
-
-    SpiceKbdInstance *keyboard;
-    SpiceMouseInstance *mouse;
-    SpiceTabletInstance *tablet;
-};
-
-typedef struct RedInputsPipeItem {
-    RedPipeItem base;
-} RedInputsPipeItem;
-
 typedef struct RedKeyModifiersPipeItem {
     RedPipeItem base;
     uint8_t modifiers;
@@ -129,8 +137,6 @@ typedef struct RedInputsInitPipeItem {
     uint8_t modifiers;
 } RedInputsInitPipeItem;
 
-static SpiceTimer *key_modifiers_timer;
-
 
 #define KEY_MODIFIERS_TTL (MSEC_PER_SEC * 2)
 
@@ -155,7 +161,7 @@ static uint8_t *inputs_channel_alloc_msg_rcv_buf(RedChannelClient *rcc,
                                                  uint16_t type,
                                                  uint32_t size)
 {
-    InputsChannel *inputs_channel = SPICE_CONTAINEROF(rcc->channel, InputsChannel, base);
+    InputsChannel *inputs_channel = INPUTS_CHANNEL(red_channel_client_get_channel(rcc));
 
     if (size > RECEIVE_BUF_SIZE) {
         spice_printerr("error: too large incoming message");
@@ -185,9 +191,9 @@ static void inputs_channel_release_msg_rcv_buf(RedChannelClient *rcc,
      ((state & SPICE_MOUSE_BUTTON_MASK_MIDDLE) ? VD_AGENT_MBUTTON_MASK : 0) |    \
      ((state & SPICE_MOUSE_BUTTON_MASK_RIGHT) ? VD_AGENT_RBUTTON_MASK : 0))
 
-static void activate_modifiers_watch(RedsState *reds)
+static void activate_modifiers_watch(InputsChannel *inputs, RedsState *reds)
 {
-    reds_core_timer_start(reds, key_modifiers_timer, KEY_MODIFIERS_TTL);
+    reds_core_timer_start(reds, inputs->key_modifiers_timer, KEY_MODIFIERS_TTL);
 }
 
 static void kbd_push_scan(SpiceKbdInstance *sin, uint8_t scan)
@@ -263,7 +269,7 @@ static void inputs_channel_send_item(RedChannelClient *rcc, RedPipeItem *base)
             red_channel_client_init_send_data(rcc, SPICE_MSG_INPUTS_MOUSE_MOTION_ACK, base);
             break;
         case RED_PIPE_ITEM_MIGRATE_DATA:
-            ((InputsChannel*)rcc->channel)->src_during_migrate = FALSE;
+            INPUTS_CHANNEL(red_channel_client_get_channel(rcc))->src_during_migrate = FALSE;
             inputs_channel_client_send_migrate_data(rcc, m, base);
             break;
         default:
@@ -276,10 +282,10 @@ static void inputs_channel_send_item(RedChannelClient *rcc, RedPipeItem *base)
 static int inputs_channel_handle_parsed(RedChannelClient *rcc, uint32_t size, uint16_t type,
                                         void *message)
 {
-    InputsChannel *inputs_channel = (InputsChannel *)rcc->channel;
-    InputsChannelClient *icc = (InputsChannelClient *)rcc;
+    InputsChannel *inputs_channel = INPUTS_CHANNEL(red_channel_client_get_channel(rcc));
+    InputsChannelClient *icc = INPUTS_CHANNEL_CLIENT(rcc);
     uint32_t i;
-    RedsState *reds = red_channel_get_server(&inputs_channel->base);
+    RedsState *reds = red_channel_get_server(RED_CHANNEL(inputs_channel));
 
     switch (type) {
     case SPICE_MSGC_INPUTS_KEY_DOWN: {
@@ -287,8 +293,9 @@ static int inputs_channel_handle_parsed(RedChannelClient *rcc, uint32_t size, ui
         if (key_down->code == CAPS_LOCK_SCAN_CODE ||
             key_down->code == NUM_LOCK_SCAN_CODE ||
             key_down->code == SCROLL_LOCK_SCAN_CODE) {
-            activate_modifiers_watch(reds);
+            activate_modifiers_watch(inputs_channel, reds);
         }
+        /* fallthrough */
     }
     case SPICE_MSGC_INPUTS_KEY_UP: {
         SpiceMsgcKeyUp *key_up = message;
@@ -362,12 +369,15 @@ static int inputs_channel_handle_parsed(RedChannelClient *rcc, uint32_t size, ui
                 reds_handle_agent_mouse_event(reds, &inputs_channel->mouse_state);
             } else if (inputs_channel_get_tablet(inputs_channel)) {
                 SpiceTabletInterface *sif;
-                sif = SPICE_CONTAINEROF(inputs_channel_get_tablet(inputs_channel)->base.sif, SpiceTabletInterface, base);
-                sif->wheel(inputs_channel_get_tablet(inputs_channel), dz, RED_MOUSE_STATE_TO_LOCAL(mouse_press->buttons_state));
+                sif = SPICE_CONTAINEROF(inputs_channel_get_tablet(inputs_channel)->base.sif,
+                                        SpiceTabletInterface, base);
+                sif->wheel(inputs_channel_get_tablet(inputs_channel), dz,
+                           RED_MOUSE_STATE_TO_LOCAL(mouse_press->buttons_state));
             }
         } else if (inputs_channel_get_mouse(inputs_channel)) {
             SpiceMouseInterface *sif;
-            sif = SPICE_CONTAINEROF(inputs_channel_get_mouse(inputs_channel)->base.sif, SpiceMouseInterface, base);
+            sif = SPICE_CONTAINEROF(inputs_channel_get_mouse(inputs_channel)->base.sif,
+                                    SpiceMouseInterface, base);
             sif->motion(inputs_channel_get_mouse(inputs_channel), 0, 0, dz,
                         RED_MOUSE_STATE_TO_LOCAL(mouse_press->buttons_state));
         }
@@ -382,12 +392,15 @@ static int inputs_channel_handle_parsed(RedChannelClient *rcc, uint32_t size, ui
                 reds_handle_agent_mouse_event(reds, &inputs_channel->mouse_state);
             } else if (inputs_channel_get_tablet(inputs_channel)) {
                 SpiceTabletInterface *sif;
-                sif = SPICE_CONTAINEROF(inputs_channel_get_tablet(inputs_channel)->base.sif, SpiceTabletInterface, base);
-                sif->buttons(inputs_channel_get_tablet(inputs_channel), RED_MOUSE_STATE_TO_LOCAL(mouse_release->buttons_state));
+                sif = SPICE_CONTAINEROF(inputs_channel_get_tablet(inputs_channel)->base.sif,
+                                        SpiceTabletInterface, base);
+                sif->buttons(inputs_channel_get_tablet(inputs_channel),
+                             RED_MOUSE_STATE_TO_LOCAL(mouse_release->buttons_state));
             }
         } else if (inputs_channel_get_mouse(inputs_channel)) {
             SpiceMouseInterface *sif;
-            sif = SPICE_CONTAINEROF(inputs_channel_get_mouse(inputs_channel)->base.sif, SpiceMouseInterface, base);
+            sif = SPICE_CONTAINEROF(inputs_channel_get_mouse(inputs_channel)->base.sif,
+                                    SpiceMouseInterface, base);
             sif->buttons(inputs_channel_get_mouse(inputs_channel),
                          RED_MOUSE_STATE_TO_LOCAL(mouse_release->buttons_state));
         }
@@ -417,7 +430,7 @@ static int inputs_channel_handle_parsed(RedChannelClient *rcc, uint32_t size, ui
             kbd_push_scan(keyboard, CAPS_LOCK_SCAN_CODE);
             kbd_push_scan(keyboard, CAPS_LOCK_SCAN_CODE | 0x80);
         }
-        activate_modifiers_watch(reds);
+        activate_modifiers_watch(inputs_channel, reds);
         break;
     }
     case SPICE_MSGC_DISCONNECTING:
@@ -462,13 +475,13 @@ static void inputs_channel_on_disconnect(RedChannelClient *rcc)
     if (!rcc) {
         return;
     }
-    inputs_release_keys(SPICE_CONTAINEROF(rcc->channel, InputsChannel, base));
+    inputs_release_keys(INPUTS_CHANNEL(red_channel_client_get_channel(rcc)));
 }
 
 static void inputs_pipe_add_init(RedChannelClient *rcc)
 {
     RedInputsInitPipeItem *item = spice_malloc(sizeof(RedInputsInitPipeItem));
-    InputsChannel *inputs = SPICE_CONTAINEROF(rcc->channel, InputsChannel, base);
+    InputsChannel *inputs = INPUTS_CHANNEL(red_channel_client_get_channel(rcc));
 
     red_pipe_item_init(&item->base, RED_PIPE_ITEM_INPUTS_INIT);
     item->modifiers = kbd_get_leds(inputs_channel_get_keyboard(inputs));
@@ -515,18 +528,18 @@ static void inputs_connect(RedChannel *channel, RedClient *client,
 
 static void inputs_migrate(RedChannelClient *rcc)
 {
-    InputsChannel *inputs = SPICE_CONTAINEROF(rcc->channel, InputsChannel, base);
+    InputsChannel *inputs = INPUTS_CHANNEL(red_channel_client_get_channel(rcc));
     inputs->src_during_migrate = TRUE;
     red_channel_client_default_migrate(rcc);
 }
 
 static void inputs_channel_push_keyboard_modifiers(InputsChannel *inputs, uint8_t modifiers)
 {
-    if (!inputs || !red_channel_is_connected(&inputs->base) ||
+    if (!inputs || !red_channel_is_connected(RED_CHANNEL(inputs)) ||
         inputs->src_during_migrate) {
         return;
     }
-    red_channel_pipes_new_add_push(&inputs->base,
+    red_channel_pipes_new_add_push(RED_CHANNEL(inputs),
         red_inputs_key_modifiers_item_new, (void*)&modifiers);
 }
 
@@ -551,8 +564,8 @@ static int inputs_channel_handle_migrate_data(RedChannelClient *rcc,
                                               uint32_t size,
                                               void *message)
 {
-    InputsChannelClient *icc = (InputsChannelClient*)rcc;
-    InputsChannel *inputs = SPICE_CONTAINEROF(rcc->channel, InputsChannel, base);
+    InputsChannelClient *icc = INPUTS_CHANNEL_CLIENT(rcc);
+    InputsChannel *inputs = INPUTS_CHANNEL(red_channel_client_get_channel(rcc));
     SpiceMigrateDataHeader *header;
     SpiceMigrateDataInputs *mig_data;
 
@@ -572,44 +585,65 @@ static int inputs_channel_handle_migrate_data(RedChannelClient *rcc,
 
 InputsChannel* inputs_channel_new(RedsState *reds)
 {
-    ChannelCbs channel_cbs = { NULL, };
+    return  g_object_new(TYPE_INPUTS_CHANNEL,
+                         "spice-server", reds,
+                         "core-interface", reds_get_core_interface(reds),
+                         "channel-type", (int)SPICE_CHANNEL_INPUTS,
+                         "id", 0,
+                         "handle-acks", FALSE,
+                         "migration-flags",
+                         (guint)(SPICE_MIGRATE_NEED_FLUSH | SPICE_MIGRATE_NEED_DATA_TRANSFER),
+                         NULL);
+
+}
+
+static void
+inputs_channel_constructed(GObject *object)
+{
     ClientCbs client_cbs = { NULL, };
-    InputsChannel *inputs;
+    InputsChannel *self = INPUTS_CHANNEL(object);
+    RedsState *reds = red_channel_get_server(RED_CHANNEL(self));
 
-    channel_cbs.config_socket = inputs_channel_config_socket;
-    channel_cbs.on_disconnect = inputs_channel_on_disconnect;
-    channel_cbs.send_item = inputs_channel_send_item;
-    channel_cbs.alloc_recv_buf = inputs_channel_alloc_msg_rcv_buf;
-    channel_cbs.release_recv_buf = inputs_channel_release_msg_rcv_buf;
-    channel_cbs.handle_migrate_data = inputs_channel_handle_migrate_data;
-    channel_cbs.handle_migrate_flush_mark = inputs_channel_handle_migrate_flush_mark;
-
-    inputs = (InputsChannel *)red_channel_create_parser(
-                                    sizeof(InputsChannel),
-                                    reds,
-                                    reds_get_core_interface(reds),
-                                    SPICE_CHANNEL_INPUTS, 0,
-                                    FALSE, /* handle_acks */
-                                    spice_get_client_channel_parser(SPICE_CHANNEL_INPUTS, NULL),
-                                    inputs_channel_handle_parsed,
-                                    &channel_cbs,
-                                    SPICE_MIGRATE_NEED_FLUSH | SPICE_MIGRATE_NEED_DATA_TRANSFER);
-
-    if (!inputs) {
-        spice_error("failed to allocate Inputs Channel");
-    }
+    G_OBJECT_CLASS(inputs_channel_parent_class)->constructed(object);
 
     client_cbs.connect = inputs_connect;
     client_cbs.migrate = inputs_migrate;
-    red_channel_register_client_cbs(&inputs->base, &client_cbs, NULL);
+    red_channel_register_client_cbs(RED_CHANNEL(self), &client_cbs, NULL);
 
-    red_channel_set_cap(&inputs->base, SPICE_INPUTS_CAP_KEY_SCANCODE);
-    reds_register_channel(reds, &inputs->base);
+    red_channel_set_cap(RED_CHANNEL(self), SPICE_INPUTS_CAP_KEY_SCANCODE);
+    reds_register_channel(reds, RED_CHANNEL(self));
 
-    if (!(key_modifiers_timer = reds_core_timer_add(reds, key_modifiers_sender, inputs))) {
+    self->key_modifiers_timer = reds_core_timer_add(reds, key_modifiers_sender, self);
+    if (!self->key_modifiers_timer) {
         spice_error("key modifiers timer create failed");
     }
-    return inputs;
+}
+
+static void
+inputs_channel_init(InputsChannel *self)
+{
+}
+
+
+static void
+inputs_channel_class_init(InputsChannelClass *klass)
+{
+    GObjectClass *object_class = G_OBJECT_CLASS(klass);
+    RedChannelClass *channel_class = RED_CHANNEL_CLASS(klass);
+
+    object_class->constructed = inputs_channel_constructed;
+
+    channel_class->parser = spice_get_client_channel_parser(SPICE_CHANNEL_INPUTS, NULL);
+    channel_class->handle_parsed = inputs_channel_handle_parsed;
+
+    /* channel callbacks */
+    channel_class->config_socket = inputs_channel_config_socket;
+    channel_class->on_disconnect = inputs_channel_on_disconnect;
+    channel_class->send_item = inputs_channel_send_item;
+    channel_class->alloc_recv_buf = inputs_channel_alloc_msg_rcv_buf;
+    channel_class->release_recv_buf = inputs_channel_release_msg_rcv_buf;
+    channel_class->handle_migrate_data = inputs_channel_handle_migrate_data;
+    channel_class->handle_migrate_flush_mark = inputs_channel_handle_migrate_flush_mark;
 }
 
 static SpiceKbdInstance* inputs_channel_get_keyboard(InputsChannel *inputs)
@@ -624,7 +658,7 @@ int inputs_channel_set_keyboard(InputsChannel *inputs, SpiceKbdInstance *keyboar
         return -1;
     }
     inputs->keyboard = keyboard;
-    inputs->keyboard->st = spice_kbd_state_new(red_channel_get_server(&inputs->base));
+    inputs->keyboard->st = spice_kbd_state_new(red_channel_get_server(RED_CHANNEL(inputs)));
     return 0;
 }
 
